@@ -45,7 +45,8 @@ enum class TreeProcessType : int {
 enum class PredictorType : int {
   kAuto = 0,
   kCPUPredictor,
-  kGPUPredictor
+  kGPUPredictor,
+  kOneAPIPredictor
 };
 }  // namespace xgboost
 
@@ -93,6 +94,7 @@ struct GBTreeTrainParam : public XGBoostParameter<GBTreeTrainParam> {
         .add_enum("auto", PredictorType::kAuto)
         .add_enum("cpu_predictor", PredictorType::kCPUPredictor)
         .add_enum("gpu_predictor", PredictorType::kGPUPredictor)
+        .add_enum("oneapi_predictor", PredictorType::kOneAPIPredictor)
         .describe("Predictor algorithm type");
     DMLC_DECLARE_FIELD(tree_method)
         .set_default(TreeMethod::kAuto)
@@ -150,6 +152,50 @@ struct DartTrainParam : public XGBoostParameter<DartTrainParam> {
   }
 };
 
+namespace detail {
+// From here on, layer becomes concrete trees.
+inline std::pair<uint32_t, uint32_t> LayerToTree(gbm::GBTreeModel const &model,
+                                                 GBTreeTrainParam const &tparam,
+                                                 size_t layer_begin,
+                                                 size_t layer_end) {
+  bst_group_t groups = model.learner_model_param->num_output_group;
+  uint32_t tree_begin = layer_begin * groups * tparam.num_parallel_tree;
+  uint32_t tree_end = layer_end * groups * tparam.num_parallel_tree;
+  if (tree_end == 0) {
+    tree_end = static_cast<uint32_t>(model.trees.size());
+  }
+  CHECK_LT(tree_begin, tree_end);
+  return {tree_begin, tree_end};
+}
+
+// Call fn for each pair of input output tree.  Return true if index is out of bound.
+template <typename Func>
+inline bool SliceTrees(int32_t layer_begin, int32_t layer_end, int32_t step,
+                       GBTreeModel const &model, GBTreeTrainParam const &tparam,
+                       uint32_t layer_trees, Func fn) {
+  uint32_t tree_begin, tree_end;
+  std::tie(tree_begin, tree_end) = detail::LayerToTree(model, tparam, layer_begin, layer_end);
+  if (tree_end > model.trees.size()) {
+    return true;
+  }
+
+  layer_end = layer_end == 0 ? model.trees.size() / layer_trees : layer_end;
+  uint32_t n_layers = (layer_end - layer_begin) / step;
+  int32_t in_it = tree_begin;
+  int32_t out_it = 0;
+  for (uint32_t l = 0; l < n_layers; ++l) {
+    for (uint32_t i = 0; i < layer_trees; ++i) {
+      CHECK_LT(in_it, tree_end);
+      fn(in_it, out_it);
+      out_it++;
+      in_it++;
+    }
+    in_it += (step - 1) * layer_trees;
+  }
+  return false;
+}
+}  // namespace detail
+
 // gradient boosted trees
 class GBTree : public GradientBooster {
  public:
@@ -198,6 +244,15 @@ class GBTree : public GradientBooster {
     return model_.learner_model_param->num_output_group == 1;
   }
 
+  // Number of trees per layer.
+  auto LayerTrees() const {
+    auto n_trees = model_.learner_model_param->num_output_group * tparam_.num_parallel_tree;
+    return n_trees;
+  }
+  // slice the trees, out must be already allocated
+  void Slice(int32_t layer_begin, int32_t layer_end, int32_t step,
+             GradientBooster *out, bool* out_of_bound) const override;
+
   void PredictBatch(DMatrix* p_fmat,
                     PredictionCacheEntry* out_preds,
                     bool training,
@@ -205,16 +260,11 @@ class GBTree : public GradientBooster {
 
   void InplacePredict(dmlc::any const &x, float missing,
                       PredictionCacheEntry *out_preds,
-                      uint32_t layer_begin = 0,
-                      unsigned layer_end = 0) const override {
+                      uint32_t layer_begin,
+                      unsigned layer_end) const override {
     CHECK(configured_);
-    // From here on, layer becomes concrete trees.
-    bst_group_t groups = model_.learner_model_param->num_output_group;
-    uint32_t tree_begin = layer_begin * groups * tparam_.num_parallel_tree;
-    uint32_t tree_end = layer_end * groups * tparam_.num_parallel_tree;
-    if (tree_end == 0 || tree_end > model_.trees.size()) {
-      tree_end = static_cast<uint32_t>(model_.trees.size());
-    }
+    uint32_t tree_begin, tree_end;
+    std::tie(tree_begin, tree_end) = detail::LayerToTree(model_, tparam_, layer_begin, layer_end);
     this->GetPredictor()->InplacePredict(x, model_, missing, out_preds,
                                          tree_begin, tree_end);
   }
@@ -228,26 +278,25 @@ class GBTree : public GradientBooster {
   }
 
   void PredictLeaf(DMatrix* p_fmat,
-                   std::vector<bst_float>* out_preds,
+                   HostDeviceVector<bst_float>* out_preds,
                    unsigned ntree_limit) override {
-    CHECK(configured_);
-    cpu_predictor_->PredictLeaf(p_fmat, out_preds, model_, ntree_limit);
+    this->GetPredictor()->PredictLeaf(p_fmat, out_preds, model_, ntree_limit);
   }
 
   void PredictContribution(DMatrix* p_fmat,
-                           std::vector<bst_float>* out_contribs,
-                           unsigned ntree_limit, bool approximate, int condition,
-                           unsigned condition_feature) override {
+                           HostDeviceVector<bst_float>* out_contribs,
+                           unsigned ntree_limit, bool approximate,
+                           int, unsigned) override {
     CHECK(configured_);
-    cpu_predictor_->PredictContribution(p_fmat, out_contribs, model_,
-                                        ntree_limit, nullptr, approximate);
+    this->GetPredictor()->PredictContribution(
+        p_fmat, out_contribs, model_, ntree_limit, nullptr, approximate);
   }
 
   void PredictInteractionContributions(DMatrix* p_fmat,
-                                       std::vector<bst_float>* out_contribs,
+                                       HostDeviceVector<bst_float>* out_contribs,
                                        unsigned ntree_limit, bool approximate) override {
     CHECK(configured_);
-    cpu_predictor_->PredictInteractionContributions(p_fmat, out_contribs, model_,
+    this->GetPredictor()->PredictInteractionContributions(p_fmat, out_contribs, model_,
                                                     ntree_limit, nullptr, approximate);
   }
 
@@ -292,6 +341,9 @@ class GBTree : public GradientBooster {
 #if defined(XGBOOST_USE_CUDA)
   std::unique_ptr<Predictor> gpu_predictor_;
 #endif  // defined(XGBOOST_USE_CUDA)
+#if defined(XGBOOST_USE_ONEAPI)
+  std::unique_ptr<Predictor> oneapi_predictor_;
+#endif  // defined(XGBOOST_USE_ONEAPI)
   common::Monitor monitor_;
 };
 
