@@ -18,11 +18,11 @@ struct LinearSquareLoss {
   // duplication is necessary, as __device__ specifier
   // cannot be made conditional on template parameter
   XGBOOST_DEVICE static bst_float PredTransform(bst_float x) { return x; }
-  XGBOOST_DEVICE static bool CheckLabel(bst_float x) { return true; }
+  XGBOOST_DEVICE static bool CheckLabel(bst_float) { return true; }
   XGBOOST_DEVICE static bst_float FirstOrderGradient(bst_float predt, bst_float label) {
     return predt - label;
   }
-  XGBOOST_DEVICE static bst_float SecondOrderGradient(bst_float predt, bst_float label) {
+  XGBOOST_DEVICE static bst_float SecondOrderGradient(bst_float, bst_float) {
     return 1.0f;
   }
   template <typename T>
@@ -72,7 +72,7 @@ struct LogisticRegression {
   XGBOOST_DEVICE static bst_float FirstOrderGradient(bst_float predt, bst_float label) {
     return predt - label;
   }
-  XGBOOST_DEVICE static bst_float SecondOrderGradient(bst_float predt, bst_float label) {
+  XGBOOST_DEVICE static bst_float SecondOrderGradient(bst_float predt, bst_float) {
     const float eps = 1e-16f;
     return fmaxf(predt * (1.0f - predt), eps);
   }
@@ -98,9 +98,40 @@ struct LogisticRegression {
   static const char* Name() { return "reg:logistic"; }
 };
 
+struct PseudoHuberError {
+  XGBOOST_DEVICE static bst_float PredTransform(bst_float x) {
+    return x;
+  }
+  XGBOOST_DEVICE static bool CheckLabel(bst_float) {
+    return true;
+  }
+  XGBOOST_DEVICE static bst_float FirstOrderGradient(bst_float predt, bst_float label) {
+    const float z = predt - label;
+    const float scale_sqrt = std::sqrt(1 + std::pow(z, 2));
+    return z/scale_sqrt;
+  }
+  XGBOOST_DEVICE static bst_float SecondOrderGradient(bst_float predt, bst_float label) {
+    const float scale = 1 + std::pow(predt - label, 2);
+    const float scale_sqrt = std::sqrt(scale);
+    return 1/(scale*scale_sqrt);
+  }
+  static bst_float ProbToMargin(bst_float base_score) {
+    return base_score;
+  }
+  static const char* LabelErrorMsg() {
+    return "";
+  }
+  static const char* DefaultEvalMetric() {
+    return "mphe";
+  }
+  static const char* Name() {
+    return "reg:pseudohubererror";
+  }
+};
+
 // logistic loss for binary classification task
 struct LogisticClassification : public LogisticRegression {
-  static const char* DefaultEvalMetric() { return "error"; }
+  static const char* DefaultEvalMetric() { return "logloss"; }
   static const char* Name() { return "binary:logistic"; }
 };
 
@@ -113,7 +144,7 @@ struct LogisticRaw : public LogisticRegression {
     predt = common::Sigmoid(predt);
     return predt - label;
   }
-  XGBOOST_DEVICE static bst_float SecondOrderGradient(bst_float predt, bst_float label) {
+  XGBOOST_DEVICE static bst_float SecondOrderGradient(bst_float predt, bst_float) {
     const float eps = 1e-16f;
     predt = common::Sigmoid(predt);
     return fmaxf(predt * (1.0f - predt), eps);
@@ -131,6 +162,9 @@ struct LogisticRaw : public LogisticRegression {
     predt = common::Sigmoid(predt);
     return std::max(predt * (T(1.0f) - predt), eps);
   }
+  static bst_float ProbToMargin(bst_float base_score) {
+    return base_score;
+  }
   static const char* DefaultEvalMetric() { return "auc"; }
 
   static const char* Name() { return "binary:logitraw"; }
